# U-Net implementation adapted from:  https://github.com/lucidrains/x-unet/blob/main/x_unet/x_unet.py
import math
from functools import partial

import torch
import torch.nn.functional as F
from beartype import beartype
from beartype.typing import Optional, Tuple, Union
from einops import rearrange, reduce, repeat
from einops.layers.torch import Rearrange
from torch import einsum, nn
from .unet import moving_normalize, log_transform, STFT

# helper functions


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


def is_power_two(n):
    return math.log2(n).is_integer()


def divisible_by(num, denom):
    return (num % denom) == 0


def cast_tuple(val, length=None):
    if isinstance(val, list):
        val = tuple(val)

    output = val if isinstance(val, tuple) else ((val,) * default(length, 1))

    if exists(length):
        assert len(output) == length

    return output


# helper classes

# def Upsample(dim, dim_out):
#     return nn.ConvTranspose3d(dim, dim_out, (1, 4, 4), (1, 2, 2), (0, 1, 1))

# def Downsample(dim, dim_out):
#     return nn.Sequential(
#         Rearrange('b c f (h s1) (w s2) -> b (c s1 s2) f h w', s1 = 2, s2 = 2),
#         nn.Conv3d(dim * 4, dim_out, 1)
#     )


def Upsample(dim, dim_out, stride=(1, 1, 4)):

    kernel_size = tuple(s * 2 if s > 1 else 1 for s in stride)
    padding = tuple(s // 2 for s in stride)

    # return nn.ConvTranspose3d(dim, dim_out, kernel_size, stride, padding, output_padding) # (1,1,7) (1,1,4) (0,0,3)
    return nn.ConvTranspose3d(dim, dim_out, kernel_size, stride, padding)


def Downsample(dim, dim_out, stride=(1, 1, 4)):

    return nn.Sequential(
        Rearrange("b c (f s1) (h s2) (w s3) -> b (c s1 s2 s3) f h w", s1=stride[0], s2=stride[1], s3=stride[2]),
        nn.Conv3d(dim * stride[0] * stride[1] * stride[2], dim_out, 1),
    )


def MergeFrequency(dim_in):
    """
    Merges frequency dimension to 1 using a linear layer.
    Args:
        dim_in: Input dimension
        dim_out: Output dimension
    Returns:
        nn.Module that transforms (b, c, s, f, t) -> (b, c, s, 1, t)
    """
    return nn.Sequential(Rearrange("b c s f t -> b c s t f"), nn.Linear(dim_in, 1), Rearrange("b c s t 1 -> b c s 1 t"))


class MergeBranch(nn.Module):
    """
    Merge two branches of the same dimension.
    Args:
        dim_in: Input dimension
    Returns:
        nn.Module that transforms (b, c, s, f, t) -> (b, c, s, 1, t)
    """

    def __init__(self, dim_in):
        super().__init__()
        self.conv = nn.Conv3d(dim_in * 2, dim_in, 1)

    def forward(self, x1, x2):
        return self.conv(torch.cat((x1, x2), dim=1))


class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x):
        return self.fn(x) + x


class RMSNorm(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.scale = dim**0.5
        self.gamma = nn.Parameter(torch.ones(dim, 1, 1, 1))

    def forward(self, x):
        return F.normalize(x, dim=1) * self.scale * self.gamma


class LayerNorm(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.gamma = nn.Parameter(torch.ones(1, dim, 1, 1, 1))

    def forward(self, x):
        eps = 1e-5 if x.dtype == torch.float32 else 1e-3
        var = torch.var(x, dim=1, unbiased=False, keepdim=True)
        mean = torch.mean(x, dim=1, keepdim=True)
        return (x - mean) / (var + eps).sqrt() * self.gamma


class WeightStandardizedConv3d(nn.Conv3d):
    def forward(self, x):
        eps = 1e-5 if x.dtype == torch.float32 else 1e-3

        weight = self.weight

        mean = reduce(weight, "o ... -> o 1 1 1 1", "mean")
        var = reduce(weight, "o ... -> o 1 1 1 1", partial(torch.var, unbiased=False))
        weight = (weight - mean) * (var + eps).rsqrt()

        return F.conv3d(x, weight, self.bias, self.stride, self.padding, self.dilation, self.groups)


# resnet blocks


class Block(nn.Module):
    def __init__(
        self,
        dim,
        dim_out,
        groups=8,
        weight_standardize=False,
        kernel_size=(1, 1, 7),
    ):
        super().__init__()
        kernel_conv_kwargs = partial(kernel_and_same_pad, kernel_size[0])
        conv = nn.Conv3d if not weight_standardize else WeightStandardizedConv3d

        self.proj = conv(dim, dim_out, **kernel_conv_kwargs(kernel_size[1], kernel_size[2]))
        self.norm = RMSNorm(dim_out)
        self.act = nn.SiLU()

    def forward(self, x):
        x = self.proj(x)
        x = self.norm(x)
        return self.act(x)


class ResnetBlock(nn.Module):
    def __init__(
        self,
        dim,
        dim_out,
        kernel_size=(1, 1, 7),
        nested_unet_depth=0,
        nested_unet_dim=32,
        weight_standardize=False,
    ):
        super().__init__()
        self.block1 = Block(
            dim,
            dim_out,
            weight_standardize=weight_standardize,
            kernel_size=kernel_size,
        )

        if nested_unet_depth > 0:
            self.block2 = NestedResidualUnet(
                dim_out,
                depth=nested_unet_depth,
                M=nested_unet_dim,
                kernel_size=kernel_size,
                weight_standardize=weight_standardize,
                add_residual=True,
            )
        else:
            self.block2 = Block(
                dim_out,
                dim_out,
                weight_standardize=weight_standardize,
                kernel_size=kernel_size,
            )

        self.res_conv = nn.Conv3d(dim, dim_out, 1) if dim != dim_out else nn.Identity()

    def forward(self, x):
        h = self.block1(x)
        h = self.block2(h)
        return h + self.res_conv(x)


# convnext 2


class GRN(nn.Module):
    """global response normalization, proposed in updated convnext paper"""

    def __init__(self, dim, eps=1e-5):
        super().__init__()
        self.eps = eps
        self.gamma = nn.Parameter(torch.zeros(dim, 1, 1, 1))
        self.bias = nn.Parameter(torch.zeros(dim, 1, 1, 1))

    def forward(self, x):
        spatial_l2_norm = x.norm(p=2, dim=(2, 3, 4), keepdim=True)
        feat_norm = spatial_l2_norm / spatial_l2_norm.mean(dim=-1, keepdim=True).clamp(min=self.eps)
        return x * feat_norm * self.gamma + self.bias + x


class ConvNextBlock(nn.Module):
    def __init__(
        self,
        dim,
        dim_out,
        *,
        mult=2,
        kernel_size=(1, 1, 7),
        nested_unet_depth=0,
        nested_unet_dim=32,
    ):
        super().__init__()
        kernel_conv_kwargs = partial(kernel_and_same_pad, kernel_size[0])

        init_kernel_size = [k + 4 if k > 1 else 1 for k in kernel_size]
        self.ds_conv = nn.Conv3d(dim, dim, **kernel_conv_kwargs(init_kernel_size[1], init_kernel_size[2]), groups=dim)

        inner_dim = dim_out * mult

        self.net = nn.Sequential(
            LayerNorm(dim),
            nn.Conv3d(dim, inner_dim, **kernel_conv_kwargs(kernel_size[1], kernel_size[2]), groups=dim_out),
            nn.GELU(),
            GRN(inner_dim),
            nn.Conv3d(inner_dim, dim_out, **kernel_conv_kwargs(kernel_size[1], kernel_size[2]), groups=dim_out),
        )

        self.nested_unet = (
            NestedResidualUnet(
                dim_out, kernel_size=kernel_size, depth=nested_unet_depth, M=nested_unet_dim, add_residual=True
            )
            if nested_unet_depth > 0
            else nn.Identity()
        )

        self.res_conv = nn.Conv3d(dim, dim_out, 1) if dim != dim_out else nn.Identity()

    def forward(self, x, time_emb=None):
        h = self.ds_conv(x)
        h = self.net(h)
        h = self.nested_unet(h)
        return h + self.res_conv(x)


# feedforward


def FeedForward(dim, mult=4.0):
    inner_dim = int(dim * mult)
    return Residual(
        nn.Sequential(
            LayerNorm(dim),
            nn.Conv3d(dim, inner_dim, 1, bias=False),
            nn.GELU(),
            LayerNorm(inner_dim),  # properly credit assign normformer
            nn.Conv3d(inner_dim, dim, 1, bias=False),
        )
    )


# attention


class Attention(nn.Module):
    def __init__(self, dim, heads=4, dim_head=64):
        super().__init__()
        self.scale = dim_head**-0.5
        self.heads = heads
        inner_dim = heads * dim_head
        self.norm = LayerNorm(dim)

        self.to_qkv = nn.Conv3d(dim, inner_dim * 3, 1, bias=False)
        self.to_out = nn.Conv3d(inner_dim, dim, 1, bias=False)

    def forward(self, x):
        f, h, w = x.shape[-3:]

        residual = x.clone()

        x = self.norm(x)

        q, k, v = self.to_qkv(x).chunk(3, dim=1)
        q, k, v = map(lambda t: rearrange(t, "b (h c) ... -> b h (...) c", h=self.heads), (q, k, v))

        q = q * self.scale
        sim = einsum("b h i d, b h j d -> b h i j", q, k)
        attn = sim.softmax(dim=-1)

        out = einsum("b h i j, b h j d -> b h i d", attn, v)

        out = rearrange(out, "b h (f x y) d -> b (h d) f x y", f=f, x=h, y=w)
        return self.to_out(out) + residual


class TransformerBlock(nn.Module):
    def __init__(self, dim, *, depth, **kwargs):
        super().__init__()
        self.attn = Attention(dim, **kwargs)
        self.ff = FeedForward(dim)

    def forward(self, x):
        x = self.attn(x)
        x = self.ff(x)
        return x


class FeatureMapConsolidator(nn.Module):
    def __init__(self, dim, *, dim_ins=tuple(), dim_outs=tuple(), resize_fmap_before=True, conv_block_fn=None):
        super().__init__()
        assert len(dim_ins) == len(dim_outs)
        self.needs_consolidating = len(dim_ins) > 0

        block_fn = default(conv_block_fn, Block)

        self.fmap_convs = nn.ModuleList([block_fn(dim_in, dim_out) for dim_in, dim_out in zip(dim_ins, dim_outs)])
        self.resize_fmap_before = resize_fmap_before

        self.final_dim_out = dim + (sum(dim_outs) if len(dim_outs) > 0 else 0)

    def resize_fmaps(self, fmaps, height, width):
        return [F.interpolate(fmap, (fmap.shape[-3], height, width)) for fmap in fmaps]

    def forward(self, x, fmaps=None):
        target_height, target_width = x.shape[-2:]

        fmaps = default(fmaps, tuple())

        if not self.needs_consolidating:
            return x

        if self.resize_fmap_before:
            fmaps = self.resize_fmaps(fmaps, target_height, target_width)

        outs = []
        for fmap, conv in zip(fmaps, self.fmap_convs):
            outs.append(conv(fmap))

        if self.resize_fmap_before:
            outs = self.resize_fmaps(outs, target_height, target_width)

        return torch.cat((x, *outs), dim=1)


# unet


def MaybeTuple(type):
    return Union[type, Tuple[type, ...]]


def kernel_and_same_pad(*kernel_size):
    paddings = tuple(map(lambda k: k // 2, kernel_size))
    return dict(kernel_size=kernel_size, padding=paddings)


def kernel_and_output_padding(*kernel_size):
    paddings = tuple(map(lambda k: k // 2, kernel_size))
    return dict(kernel_size=kernel_size, padding=paddings, output_padding=paddings)


class XUnet(nn.Module):

    @beartype
    def __init__(
        self,
        dim,
        init_dim=None,
        out_dim=None,
        kernel_size=(1, 1, 7),
        dim_mults: MaybeTuple(int) = (1, 2, 4, 4),
        num_blocks_per_stage: MaybeTuple(int) = (2, 2, 2, 2),
        num_self_attn_per_stage: MaybeTuple(int) = (0, 0, 0, 1),
        nested_unet_depths: MaybeTuple(int) = (0, 0, 0, 0),
        nested_unet_dim=32,
        channels=3,
        use_convnext=False,
        consolidate_upsample_fmaps=True,
        skip_scale=2**-0.5,
        weight_standardize=False,
        attn_heads: MaybeTuple(int) = 8,
        attn_dim_head: MaybeTuple(int) = 32,
        moving_norm=(1024, 256),
        log_scale=False,
        add_stft=False,
        add_polarity=False,
        add_event=False,
        add_prompt=False,
    ):
        super().__init__()

        self.kernel_size = kernel_size
        self.moving_norm = moving_norm
        self.log_scale = log_scale
        self.add_stft = add_stft
        self.add_polarity = add_polarity
        self.add_event = add_event
        self.add_prompt = add_prompt

        self.train_as_images = kernel_size[0] == 1

        self.skip_scale = skip_scale
        self.channels = channels

        init_dim = default(init_dim, dim)
        init_kernel_size = [k + 4 if k > 1 else 1 for k in kernel_size]
        self.init_conv = nn.Conv3d(channels, init_dim, **kernel_and_same_pad(*init_kernel_size))

        dims = [init_dim, *map(lambda m: dim * m, dim_mults)]
        in_out = list(zip(dims[:-1], dims[1:]))

        self.scale_factor = [tuple(map(lambda k: k // 2 + 1, kernel_size))] * 4

        self.downs = nn.ModuleList([])
        self.ups = nn.ModuleList([])

        num_resolutions = len(in_out)

        # resnet or convnext

        blocks = (
            partial(
                ConvNextBlock,
                kernel_size=kernel_size,
            )
            if use_convnext
            else partial(
                ResnetBlock,
                weight_standardize=weight_standardize,
                kernel_size=kernel_size,
            )
        )

        # whether to use nested unet, as in unet squared paper
        nested_unet_depths = cast_tuple(nested_unet_depths, num_resolutions)

        # number of blocks per stage

        num_blocks_per_stage = cast_tuple(num_blocks_per_stage, num_resolutions)
        assert all([num_blocks > 0 for num_blocks in num_blocks_per_stage])

        # number of self attention blocks per stage

        num_self_attn_per_stage = cast_tuple(num_self_attn_per_stage, num_resolutions)
        assert all([num_self_attn_blocks >= 0 for num_self_attn_blocks in num_self_attn_per_stage])

        # attn kwargs

        attn_heads = cast_tuple(attn_heads, num_resolutions)
        attn_dim_head = cast_tuple(attn_dim_head, num_resolutions)

        # modules for all layers

        skip_dims = []

        down_stage_parameters = [
            in_out,
            nested_unet_depths,
            num_blocks_per_stage,
            num_self_attn_per_stage,
            attn_heads,
            attn_dim_head,
        ]

        up_stage_parameters = [reversed(params[:-1]) for params in down_stage_parameters]

        # downs

        for ind, ((dim_in, dim_out), nested_unet_depth, num_blocks, self_attn_blocks, heads, dim_head) in enumerate(
            zip(*down_stage_parameters)
        ):
            is_last = ind >= (num_resolutions - 1)
            skip_dims.append(dim_in)

            self.downs.append(
                nn.ModuleList(
                    [
                        blocks(dim_in, dim_in, nested_unet_depth=nested_unet_depth, nested_unet_dim=nested_unet_dim),
                        nn.ModuleList(
                            [
                                blocks(
                                    dim_in, dim_in, nested_unet_depth=nested_unet_depth, nested_unet_dim=nested_unet_dim
                                )
                                for _ in range(num_blocks - 1)
                            ]
                        ),
                        nn.ModuleList(
                            [
                                TransformerBlock(dim_in, depth=self_attn_blocks, heads=heads, dim_head=dim_head)
                                for _ in range(self_attn_blocks)
                            ]
                        ),
                        Downsample(
                            dim_in,
                            dim_out,
                            stride=self.scale_factor[ind],
                        ),
                    ]
                )
            )

        # middle

        mid_dim = dims[-1]
        mid_nested_unet_depth = nested_unet_depths[-1]

        self.mid = blocks(mid_dim, mid_dim, nested_unet_depth=mid_nested_unet_depth, nested_unet_dim=nested_unet_dim)
        self.mid_attn = Attention(mid_dim, heads=attn_heads[-1], dim_head=attn_dim_head[-1])
        self.mid_after = blocks(
            mid_dim, mid_dim, nested_unet_depth=mid_nested_unet_depth, nested_unet_dim=nested_unet_dim
        )

        self.mid_upsample = Upsample(
            mid_dim,
            dims[-2],
            stride=self.scale_factor[-1],
        )

        # ups

        for ind, ((dim_in, dim_out), nested_unet_depth, num_blocks, self_attn_blocks, heads, dim_head) in enumerate(
            zip(*up_stage_parameters)
        ):
            is_last = ind >= (num_resolutions - 1)

            self.ups.append(
                nn.ModuleList(
                    [
                        blocks(
                            dim_out + skip_dims.pop(),
                            dim_out,
                            nested_unet_depth=nested_unet_depth,
                            nested_unet_dim=nested_unet_dim,
                        ),
                        nn.ModuleList(
                            [
                                blocks(
                                    dim_out,
                                    dim_out,
                                    nested_unet_depth=nested_unet_depth,
                                    nested_unet_dim=nested_unet_dim,
                                )
                                for _ in range(num_blocks - 1)
                            ]
                        ),
                        nn.ModuleList(
                            [
                                TransformerBlock(dim_out, depth=self_attn_blocks, heads=heads, dim_head=dim_head)
                                for _ in range(self_attn_blocks)
                            ]
                        ),
                        (
                            Upsample(
                                dim_out,
                                dim_in,
                                stride=self.scale_factor[ind],
                            )
                            if not is_last
                            else nn.Identity()
                        ),
                    ]
                )
            )

        out_dim = default(out_dim, channels)

        if consolidate_upsample_fmaps:
            self.consolidator = FeatureMapConsolidator(
                dim,
                dim_ins=tuple(map(lambda m: dim * m, dim_mults)),
                dim_outs=(dim,) * len(dim_mults),
                conv_block_fn=blocks,
            )
        else:
            self.consolidator = FeatureMapConsolidator(dim=dim)

        final_dim_in = self.consolidator.final_dim_out

        self.final_conv = nn.Sequential(
            blocks(final_dim_in + dim, max(dim, out_dim)),
            nn.Conv3d(max(dim, out_dim), out_dim, **kernel_and_same_pad(*kernel_size)),
        )

        ## Polarity
        if self.add_polarity:
            nested_unet_depth = nested_unet_depths[0]
            num_blocks = num_blocks_per_stage[0]
            self_attn_blocks = num_self_attn_per_stage[0]
            heads = attn_heads[0]
            dim_head = attn_dim_head[0]
            dim_in = dims[0]
            init_kernel_size = [k + 4 if k > 1 else 1 for k in kernel_size]
            self.polarity_init = nn.Conv3d(1, init_dim, **kernel_and_same_pad(*init_kernel_size))
            self.polarity_encoder = nn.ModuleList([])
            self.polarity_encoder.append(
                nn.ModuleList(
                    [
                        blocks(dim_in, dim_in, nested_unet_depth=nested_unet_depth, nested_unet_dim=nested_unet_dim),
                        nn.ModuleList(
                            [
                                blocks(
                                    dim_in, dim_in, nested_unet_depth=nested_unet_depth, nested_unet_dim=nested_unet_dim
                                )
                                for _ in range(num_blocks - 1)
                            ]
                        ),
                        nn.ModuleList(
                            [
                                TransformerBlock(dim_in, depth=self_attn_blocks, heads=heads, dim_head=dim_head)
                                for _ in range(self_attn_blocks)
                            ]
                        ),
                    ]
                )
            )
            self.polarity_final = nn.Sequential(
                blocks(dim_in * 2, dim_in, nested_unet_depth=nested_unet_depth, nested_unet_dim=nested_unet_dim),
                nn.Conv3d(dim_in, out_dim, **kernel_and_same_pad(*kernel_size)),
            )

        ## Event
        if self.add_event:
            self.event_feature_level = 3
            level = self.event_feature_level
            nested_unet_depth = nested_unet_depths[level]
            num_blocks = num_blocks_per_stage[level]
            self_attn_blocks = num_self_attn_per_stage[level]
            heads = attn_heads[level]
            dim_head = attn_dim_head[level]
            dim_in = dims[level]
            self.event_final = nn.Sequential(
                blocks(dim_in, dim_in, nested_unet_depth=nested_unet_depth, nested_unet_dim=nested_unet_dim),
                Upsample(
                    dim_in,
                    dim_in,
                    stride=self.scale_factor[-1],
                    # **kernel_and_output_padding(*kernel_size),
                ),
                nn.Conv3d(dim_in, out_dim, **kernel_and_same_pad(*kernel_size)),
            )

        ## STFT
        if self.add_stft:
            self.stft = STFT(n_fft=64 + 1, hop_length=self.scale_factor[0][-1])
            self.kernel_size_stft = [kernel_size[0], 3, kernel_size[2]]
            self.spec_init = nn.Sequential(
                nn.Conv3d(
                    channels,
                    init_dim,
                    **kernel_and_same_pad(*self.kernel_size_stft),
                ),
            )
            blocks_stft = (
                partial(
                    ConvNextBlock,
                    kernel_size=self.kernel_size_stft,
                )
                if use_convnext
                else partial(
                    ResnetBlock,
                    weight_standardize=weight_standardize,
                    kernel_size=self.kernel_size_stft,
                )
            )
            self.spec_down = nn.ModuleList([])
            for ind, ((dim_in, dim_out), nested_unet_depth, num_blocks, self_attn_blocks, heads, dim_head) in enumerate(
                zip(*down_stage_parameters)
            ):
                if ind == 0:
                    continue
                self.spec_down.append(
                    nn.ModuleList(
                        [
                            blocks_stft(
                                dim_in, dim_in, nested_unet_depth=nested_unet_depth, nested_unet_dim=nested_unet_dim
                            ),
                            nn.ModuleList(
                                [
                                    blocks_stft(
                                        dim_in,
                                        dim_in,
                                        nested_unet_depth=nested_unet_depth,
                                        nested_unet_dim=nested_unet_dim,
                                    )
                                    for _ in range(num_blocks - 1)
                                ]
                            ),
                            Downsample(dim_in, dim_out, stride=self.scale_factor[ind]),
                            MergeFrequency(32),
                            MergeBranch(dim_in),
                        ]
                    )
                )
            self.mid_stft = nn.Sequential(
                blocks_stft(mid_dim, mid_dim, nested_unet_depth=mid_nested_unet_depth, nested_unet_dim=nested_unet_dim),
                MergeFrequency(32),
            )
            self.mid_merge = MergeBranch(mid_dim)

    def forward(self, x):

        is_image = x.ndim == 4

        # validations

        assert not (
            is_image and not self.train_as_images
        ), "you specified a frame kernel size for the convolutions in this unet, but you are passing in images"
        assert not (
            not is_image and self.train_as_images
        ), "you specified no frame kernel size dimension, yet you are passing in a video. fold the frame dimension into the batch"

        # cast images to 1 framed video

        if is_image:
            x = rearrange(x, "b c h w -> b c 1 h w")

        x = moving_normalize(x, filter=self.moving_norm[0], stride=self.moving_norm[1])
        if self.log_scale:
            x = log_transform(x)

        # origin
        x_origin = x.clone()

        # initial convolution

        x = self.init_conv(x)

        if self.add_polarity:
            x_polarity = self.polarity_init(x_origin[:, -1:, :, :, :])
            # x_polarity = self.polarity_init(torch.clamp(x_origin[:, -1:, :, :, :], -1.0, 1.0))

        # residual

        r = x.clone()

        # downs and ups

        down_hiddens = []
        up_hiddens = []

        for init_block, blocks, attn_blocks, downsample in self.downs:
            x = init_block(x)

            for block in blocks:
                x = block(x)

            for attn_block in attn_blocks:
                x = attn_block(x)

            down_hiddens.append(x)
            x = downsample(x)

        if self.add_stft:
            nb, nc, nx, nf, nt = x_origin.shape
            x_stft = rearrange(x_origin, "b c x f t -> (b x f) c t")  # (b x f), c, t
            x_stft = self.stft(x_stft)  # (b x f), c, nf, t
            x_stft = rearrange(x_stft, "(b x f) c nf t -> b c x (f nf) t", b=nb, x=nx, f=nf)  # b, x, (f*nf), c, t
            # if self.training:
            if True:
                sgram = x_stft.clone()
            else:
                sgram = None
            x_stft = self.spec_init(x_stft)
            for i, (init_block, blocks, downsample, merge_freq, merge_branch) in enumerate(self.spec_down):
                x_stft = init_block(x_stft)

                for block in blocks:
                    x_stft = block(x_stft)

                down_hiddens[i + 1] = merge_branch(down_hiddens[i + 1], merge_freq(x_stft))
                x_stft = downsample(x_stft)
            x_stft = self.mid_stft(x_stft)

        x = self.mid(x)
        if self.add_stft:
            x = self.mid_merge(x, x_stft)
        x = self.mid_attn(x) + x
        x = self.mid_after(x)

        if self.add_prompt:
            out_prompt = x.clone()
        else:
            out_prompt = None

        up_hiddens.append(x)
        x = self.mid_upsample(x)

        for init_block, blocks, attn_blocks, upsample in self.ups:
            x = torch.cat((x, down_hiddens.pop() * self.skip_scale), dim=1)

            x = init_block(x)

            for block in blocks:
                x = block(x)

            for attn_block in attn_blocks:
                x = attn_block(x)

            up_hiddens.insert(0, x)
            x = upsample(x)

        if self.add_polarity:
            x_polarity_aux = x.clone()

        # consolidate feature maps

        x = self.consolidator(x, up_hiddens)

        # final residual

        x = torch.cat((x, r), dim=1)
        # x = torch.cat((x, down_hiddens.pop()), dim=1)

        # final convolution

        out_phase = self.final_conv(x)

        # polarity
        if self.add_polarity:
            for init_block, blocks, attn_blocks in self.polarity_encoder:
                x_polarity = init_block(x_polarity)
                for block in blocks:
                    x_polarity = block(x_polarity)
                for attn_block in attn_blocks:
                    x_polarity = attn_block(x_polarity)

            x_polarity = torch.cat((x_polarity, x_polarity_aux), dim=1)
            out_polarity = self.polarity_final(x_polarity)
        else:
            out_polarity = None

        # event
        if self.add_event:
            x_event = up_hiddens[self.event_feature_level - 1].clone()
            out_event = self.event_final(x_event)
        else:
            out_event = None

        if is_image:
            out_phase = rearrange(out_phase, "b c 1 h w -> b c h w")
            if self.add_polarity:
                out_polarity = rearrange(out_polarity, "b c 1 h w -> b c h w")
            if self.add_event:
                out_event = rearrange(out_event, "b c 1 h w -> b c h w")
            if self.add_prompt:
                out_prompt = rearrange(out_prompt, "b c 1 h w -> b c h w")

        out = {"phase": out_phase, "polarity": out_polarity, "event": out_event, "prompt": out_prompt}
        # if self.add_stft and self.training:
        if self.add_stft:
            out["spectrogram"] = sgram.squeeze(2)  ## nb, nc, nx, nf, nt -> nb, nc, nf, nt

        return out


# RSU


class PixelShuffleUpsample(nn.Module):
    def __init__(
        self,
        dim,
        dim_out=None,
        # scale_factor = 2
        scale_factor=(1, 1, 4),
    ):
        super().__init__()
        # self.scale_squared = scale_factor ** 2
        self.scale_squared = scale_factor[0] * scale_factor[1] * scale_factor[2]
        dim_out = default(dim_out, dim)
        conv = nn.Conv3d(dim, dim_out * self.scale_squared, 1)

        # self.net = nn.Sequential(
        #     conv,
        #     nn.SiLU(),
        #     Rearrange('b (c r s) f h w -> b c f (h r) (w s)', r = scale_factor, s = scale_factor)
        # )

        self.net = nn.Sequential(
            conv,
            nn.SiLU(),
            Rearrange(
                "b (c r s t) f h w -> b c (f r) (h s) (w t)",
                r=scale_factor[0],
                s=scale_factor[1],
                t=scale_factor[2],
            ),
        )

        self.init_conv_(conv)

    def init_conv_(self, conv):
        o, i, *rest_dims = conv.weight.shape
        conv_weight = torch.empty(o // self.scale_squared, i, *rest_dims)
        nn.init.kaiming_uniform_(conv_weight)
        conv_weight = repeat(conv_weight, "o ... -> (o r) ...", r=self.scale_squared)

        conv.weight.data.copy_(conv_weight)
        nn.init.zeros_(conv.bias.data)

    def forward(self, x):
        x = self.net(x)
        return x


class NestedResidualUnet(nn.Module):
    def __init__(
        self,
        dim,
        *,
        depth,
        M=32,
        kernel_size=(1, 1, 7),
        add_residual=False,
        skip_scale=2**-0.5,
        weight_standardize=False,
    ):
        super().__init__()

        self.depth = depth
        self.downs = nn.ModuleList([])
        self.ups = nn.ModuleList([])
        self.kernel_size = kernel_size
        scale_factor = tuple([k // 2 + 1 for k in kernel_size])

        conv = WeightStandardizedConv3d if weight_standardize else nn.Conv3d

        for ind in range(depth):
            is_first = ind == 0
            dim_in = dim if is_first else M

            # down = nn.Sequential(
            #     conv(dim_in, M, (1, 4, 4), stride = (1, 2, 2), padding = (0, 1, 1)),
            #     RMSNorm(M),
            #     nn.SiLU()
            # )

            down = nn.Sequential(
                conv(
                    dim_in,
                    M,
                    kernel_size=tuple(k * 2 if k > 1 else 1 for k in scale_factor),
                    stride=scale_factor,
                    padding=tuple(k // 2 for k in scale_factor),
                ),
                RMSNorm(M),
                nn.SiLU(),
            )

            up = nn.Sequential(PixelShuffleUpsample(2 * M, dim_in, scale_factor), RMSNorm(dim_in), nn.SiLU())

            self.downs.append(down)
            self.ups.append(up)

        self.mid = nn.Sequential(
            conv(M, M, **kernel_and_same_pad(*kernel_size)),
            RMSNorm(M),
            nn.SiLU(),
        )

        self.skip_scale = skip_scale
        self.add_residual = add_residual

    def forward(self, x, residual=None):
        is_video = x.ndim == 5

        if self.add_residual:
            residual = default(residual, x.clone())

        *_, h, w = x.shape

        layers = len(self.ups)

        # for dim_name, size in (("height", h), ("width", w)):
        #     assert divisible_by(
        #         size, 2**layers
        #     ), f"{dim_name} dimension {size} must be divisible by {2 ** layers} ({layers} layers in nested unet)"
        #     assert (
        #         size % (2**self.depth)
        #     ) == 0, f"the unet has too much depth for the image {dim_name} ({size}) being passed in"
        assert divisible_by(
            w, 2**layers
        ), f"width dimension {w} must be divisible by {2 ** layers} ({layers} layers in nested unet)"
        assert (w % (2**self.depth)) == 0, f"the unet has too much depth for the image width ({w}) being passed in"
        if h > 1:
            assert divisible_by(
                h, 2**layers
            ), f"height dimension {h} must be divisible by {2 ** layers} ({layers} layers in nested unet)"
            assert (h % (2**self.depth)) == 0, f"the unet has too much depth for the image height ({h}) being passed in"

        # hiddens

        hiddens = []

        # unet

        for down in self.downs:
            x = down(x)
            hiddens.append(x.clone().contiguous())

        x = self.mid(x)

        for up in reversed(self.ups):
            x = torch.cat((x, hiddens.pop() * self.skip_scale), dim=1)
            x = up(x)

        # adding residual

        if self.add_residual:
            x = x + residual
            x = F.silu(x)

        return x


# example

if __name__ == "__main__":

    from torchinfo import summary
    from torchviz import make_dot

    model = XUnet(
        dim=8,
        out_dim=16,
        dim_mults=(1, 2, 4, 8),
        add_polarity=True,
        add_event=True,
        add_stft=True,
    )

    # print(model)
    data = torch.randn(7, 3, 1, 4096)

    summary(model, input_size=data.shape, depth=5)

    model.to("cpu")
    out = model(data)

    dot = make_dot(out["phase"], params=dict(model.named_parameters()))
    dot.render("x_unet", format="png")

    for k, v in out.items():
        if v is not None:
            print(f"{k}: {v.shape}")
